from .ddpg import DDPG
from .dqn import DQN
from .fqf import FQF
from .iqn import IQN
from .misc import SlacObservation
from .ppo import PPO
from .qrdqn import QRDQN
from .sac import SAC
from .sac_ae import SAC_AE
from .sac_discor import SAC_DisCor
from .sac_discrete import SAC_Discrete
from .slac import SLAC
from .td3 import TD3
from .tqc import TQC
